import argparse
import pandas
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.cm
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
#import scipy as sp
import re
from sklearn.mixture import DPGMM
from sklearn.decomposition import PCA

from collections import namedtuple

#import profile
import asmparse
import collate

parser = argparse.ArgumentParser(description='Report the blocks that vary the most across runs')
parser.add_argument('inpath', help='path to collated experiment datafile')

re_prefunc_bb = re.compile(r'^[^:]*:(.*)$')

def strip_prefunc(bb_name):
    if len(bb_name) == 0:
        return ''
    return re_prefunc_bb.match(bb_name).group(1)

def mean_profile(panel):
    return panel.fillna(0).mean(axis=0)

def analyze_app(active_prof, working_prof, asm_file,
                cols=['line', 'col', 'bb', 'contention', 'active_instability', 'active_total', 'working_instability', 'working_total'],
                sort=['line', 'col']):
    asm_df = asmparse.codepoints_dataframe(asm_file)
    df = unstable_contending_block(active_prof, working_prof)
    df.index = map(strip_prefunc, df.index)
    merged = pandas.merge(asm_df, df, left_on='bb', right_index=True)
#    merged = merged.reindex(index=merged.index, columns=cols, copy=False)
    merged = merged[cols]
    merged = merged.sort(sort)
    return merged

def function_from_bb(bb):
    idx = bb.find(':')
    if idx < 0:
        function_from_bb.uuid += 1
        return str(function_from_bb.uuid)
    return bb[:idx]
function_from_bb.uuid = 0

def scale_func_szs(func_szs, sz_range=(10,60)):
    max_afs, min_afs = np.max(func_szs), np.min(func_szs)
    d_afs = max_afs - min_afs

    lb, ub = sz_range
    dx = ub - lb

    scaled_fs = lb + dx*((func_szs - min_afs) / d_afs)
    scaled_fs[np.isnan(scaled_fs)] = lb

    return scaled_fs

def pca_df(df):
    pca = PCA(n_components=2)
    X = pca.fit_transform(df)

    if np.any(np.isnan(pca.explained_variance_ratio_)):
        print pca.explained_variance_ratio_
        print df
        print X

    print "Explained variance ratio:", pca.explained_variance_ratio_
    return X

def plot_contention_instability(name, analyzed_df, sz_range=(10,60),
                                figsize=(20,8), cm_name='jet'):

    analyzed_df['func'] = pandas.Series([ function_from_bb(bb) for bb in analyzed_df.index ],
                                        index=analyzed_df.index)
    by_func = analyzed_df.groupby('func')

    fig = plt.figure(figsize=figsize)
    ax_active  = fig.add_subplot(121)
    ax_working = fig.add_subplot(122, sharey=ax_active)

    ax_active.set_xlabel('Instability')
    ax_active.set_ylabel('Contention')
    ax_active.set_title(name + ' - Active')

    ax_working.set_xlabel('Instability')
    ax_working.set_ylabel('Contention')
    ax_working.set_title(name + ' - Working')

    cm = matplotlib.cm.get_cmap(cm_name, len(by_func))

    Ngroups = len(by_func)
    a_func_szs = np.empty(Ngroups)
    w_func_szs = np.empty(Ngroups)
    for i, (_, idxs) in enumerate(by_func.groups.iteritems()):
        sub_df = analyzed_df.ix[idxs]

        # smooth
        a_func_szs[i] = np.log(sub_df['active_total'].sum() + 1)
        w_func_szs[i] = np.log(sub_df['working_total'].sum() + 1)

    scaled_afs, scaled_wfs = scale_func_szs(a_func_szs, sz_range), scale_func_szs(w_func_szs, sz_range)

    for i, (func, idxs) in enumerate(by_func.groups.iteritems()):
        sub_df = analyzed_df.ix[idxs]

        c = cm(i)

        ax_active.scatter( sub_df['active_instability'],  sub_df['contention'], s=scaled_afs[i], c=c)
        ax_active.set_xlim([0, 1.0])
        ax_active.set_ylim([0, 5.0])
        ax_working.scatter(sub_df['working_instability'], sub_df['contention'], s=scaled_wfs[i], c=c)
        ax_working.set_xlim([0, 1.0])
        ax_working.set_ylim([0, 5.0])

    return fig

def plot_pca(name, a_df, w_df, figsize=(20, 8), cm_name='jet'):
    fig = plt.figure(figsize=figsize)
    axa = fig.add_subplot(121)
    axw = fig.add_subplot(122, sharey=axa)

    axa.set_title('Active PCA')
    axw.set_title('Working PCA')

    cm = matplotlib.cm.get_cmap(cm_name, len(a_df.index))

    a_df['func'] = pandas.Series( [ function_from_bb(strip_prefunc(bb)) for
                                    bb in a_df.index ], index=a_df.index )
    by_func = a_df.groupby('func')
    del a_df['func']

    Nbb = len(a_df.index)
    afszs = np.empty(Nbb)
    wfszs = np.empty(Nbb)
    for i, (_, idxs) in enumerate(by_func.groups.iteritems()):
        sa_df, sw_df = a_df.ix[idxs], w_df.ix[idxs]
#        print sa_df.as_matrix().sum()
        afszs[i] = np.log(sa_df.as_matrix().sum() + 1)
        wfszs[i] = np.log(sw_df.as_matrix().sum() + 1)

    scaled_afs, scaled_wfs = scale_func_szs(afszs), scale_func_szs(wfszs)

    for i, (func, idxs) in enumerate(by_func.groups.iteritems()):
        sa_df, sw_df = a_df.ix[idxs], w_df.ix[idxs]

        c = cm(i)

        if np.sum(sa_df.as_matrix()) == 0.0 or np.sum(sw_df.as_matrix()) == 0.0:
            continue

        a_pca = pca_df(sa_df)
        w_pca = pca_df(sw_df)

        axa.scatter(a_pca[0], a_pca[1], s=scaled_afs[i], c=c)
        axw.scatter(w_pca[0], w_pca[1], s=scaled_wfs[i], c=c)

    return fig

def write_plots(plotter, iterable, filename):
#    for it in iterable:
#        fig = plotter(it)
#        #fig.canvas = FigureCanvasPdf(fig)
#        pp.savefig(fig)
#        fig.close()
#        print "Done with", it
#
#    pp.close()
    matplotlib.use('Agg')
    for it in iterable:
        print "Now plotting", it
        fig = plotter(it)
        plt.savefig(it, format='png')
        plt.close()


def blocks_by_variance(panel):
    # figure this one out
    pass


# 8/14: After discussion, correct normalization should be column-wise.
def matrix_for_block(panel, bb, normalize=False):
    X = panel.major_xs(bb).astype('float64').as_matrix().T
    if not normalize:
        return X
    # Now rows (axis=0) are the basic blocks and columns (axis=1)
    X_norm_na = X / X.sum(axis=0)
    return np.nan_to_num(X_norm_na)

def bbs_by_component_mean(panel):
    n_runs = len(panel.items)
    bbs = panel.major_axis
    max_threads = len(panel.major_axis)
    mean = np.zeros((len(bbs)), dtype='float64')
    denom = n_runs * max_threads
    for i, bb in enumerate(bbs):
        X = matrix_for_block(panel, bb)
        mean[i] = np.sum(X) / denom

    idxs = np.argsort(mean)[::-1]
    return bbs[idxs], mean[idxs]

def mean_var_for_block(panel):
    """Unnormalized mean and variance of blocks"""
    bbs = panel.major_axis

    summary = {}

    for i, bb in enumerate(bbs):
        X = matrix_for_block(panel, bb)
        # Averaging out dimension 0 preserves the profile
        mean = np.mean(X, 0)
        cov  = np.cov(X.T)
        summary[bb] = { 'mean': mean, 'cov': cov }

    return summary

# This should really be the sum of earth mover distance of the whole matrix
# (expressed as a vector?).
def contention_for_block(actives, workings):
    bbs = actives.major_axis
    assert(all(bbs == workings.major_axis))

    # Not all runs may be present on both sides, so we take
    # intersection
    runs = actives.items & workings.items
    a_sub, w_sub = ( actives.ix[runs].fillna(0),
                     workings.ix[runs].fillna(0) )

    # For now, the contention is the matrix norm.
    c = np.zeros(len(bbs))
    for i, bb in enumerate(bbs):
        Xa, Xw = matrix_for_block(a_sub, bb, True), matrix_for_block(w_sub, bb, True)
        Xa_rows, Xa_cols = Xa.shape
        Xw_rows, Xw_cols = Xw.shape

        # The subindexing by runs takes care of this
        assert(Xa_rows == Xw_rows)

        if Xa_cols < Xw_cols:
            Xa = np.copy(Xa)
            Xa.resize((Xa_rows, Xw_cols))
        elif Xa_cols > Xw_cols:
            Xw = np.copy(Xw)
            Xw.resize((Xa_rows, Xa_cols))

        c[i] = np.linalg.norm(Xa - Xw)

    return pandas.Series(c, index=bbs)


# TODO: Change to generator/iterator. Then you can fold the top in too!
def for_each_block(panel, f, normalize=False):
    bbs = panel.major_axis
    panel_full = panel.fillna(0)
    v = np.zeros(len(bbs))
    for i, bb in enumerate(bbs):
        X = matrix_for_block(panel_full, bb, normalize)
        v[i] = f(X)

    return pandas.Series(v, index=bbs)

def block_variance(X):
    n_exec = np.sum(X[0])
    if n_exec == 0.0: return 0.0

    S = np.cov(X.T)
    if np.all(S == 0.0): return 0.0

    d = np.diag(S)
    return np.sum(d)

def unstable_contending_block(actives, workings):
    df = pandas.DataFrame({'contention':          contention_for_block(actives, workings),
                           'active_instability':  np.sqrt(for_each_block(actives, block_variance, normalize=True)),
                           'working_instability': np.sqrt(for_each_block(workings, block_variance, normalize=True)),
                           'active_total': for_each_block(actives, np.sum),
                           'working_total': for_each_block(workings, np.sum),
                           })
    return df

# TODO: Finish these!
def cluster_block(panel, bb):
    X = matrix_for_block(panel, bb)
    model = DPGMM(n_components=30, covariance_type='diag', alpha=1, n_iter=100)
    model.fit(X)
    return model

#    # concatenate
#    X =
#    # figure this

if __name__ == "__main__":
    args = parser.parse_args()
    coll = collate.dezick(args.inpath)

