#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>

#ifdef GZIP_PROFILE
#include <zlib.h>
#endif

#ifndef PROF_MAX_THREADS
#define PROF_MAX_THREADS 16
#endif

#include <string.h>
#include <errno.h>
#define DIE_IF(exp) do { if(exp) { fprintf(stderr, "System call in instrumented code failed at %s:%d. %s: %s\n", __FILE__, __LINE__, #exp, strerror(errno)); abort(); } } while (0)
#define DIE_NZ(p)   DIE_IF((p) != 0)
#define DIE_NULL(p) DIE_IF((p) == NULL)

// Basic block names
extern char *PROF_bbmap[];
extern const int PROF_nbbs;

// Size calculations, in terms of uint32_t
// Everything defined in terms of PROF_nbbs and PROF_MAX_THREADS
#define PER_THREADCOUNT_N PROF_nbbs
#define PER_THREAD_N (PROF_MAX_THREADS * PER_THREADCOUNT_N)
#define PROFILE_N (PROF_MAX_THREADS * PER_THREAD_N)

// Atexit handlers
static pthread_key_t k_thread_exit;

// Histograms
static uint32_t *PROF_hist;
static  __thread uint32_t *PROF_thread_hist;

// volatile pointer to nonvolatile
typedef uint32_t * volatile atomic_ptr_uint32_t __attribute__ ((aligned(8)));
typedef volatile int32_t atomic_int32_t __attribute ((aligned(8)));
static atomic_ptr_uint32_t PROF_curr_thread_hist;
static atomic_int32_t PROF_n_offset;

#define PROF_INC_OFFSET() __sync_add_and_fetch(&PROF_n_offset, PER_THREADCOUNT_N)
#define PROF_DEC_OFFSET() __sync_sub_and_fetch(&PROF_n_offset, PER_THREADCOUNT_N)

#define PROF_INC_CURR() __sync_add_and_fetch(&PROF_curr_thread_hist, (atomic_ptr_uint32_t)(sizeof(uint32_t) * PER_THREAD_N))
#define PROF_DEC_CURR() __sync_sub_and_fetch(&PROF_curr_thread_hist, (atomic_ptr_uint32_t)(sizeof(uint32_t) * PER_THREAD_N))

struct PROF_wrapper_t {
  void *(*start_routine)(void*);
  void *arg;
};

// Timing
static struct tm start_time;

// Atexit handler functions
static void handle_thread_exit(void *p) {
  fprintf(stderr, "[%lu] handle_thread_exit\n", pthread_self());
  /*
  PROF_DEC_OFFSET();
  PROF_DEC_CURR();
  fprintf(stderr, "[%lu] handle_thread_exit completed; PROF_n_offset = %x; PROF_curr_thread_hist = %x\n",
          pthread_self(), PROF_n_offset, PROF_curr_thread_hist);
  */
}

// ====================
// WRITING OUT TO FILE
// ====================
static void merge_hists() {
  int i, t;
  uint32_t *curr_hist;
  for (t = 1; t < PROF_MAX_THREADS; t++) {
    curr_hist = PROF_hist + t*PER_THREAD_N;
    for (i = 0; i < PER_THREAD_N; i++) {
      PROF_hist[i] += curr_hist[i];
    }
  }
}

#define DATE_LEN 16
#define OUTPATH_LEN 512
static void out_filename(char *buf, int buflen, const char *basename) {
  static char empty[] = "";
  char datebuf[DATE_LEN];
  strftime(datebuf, DATE_LEN, "%m-%d_%H:%M:%S", &start_time);
  char *outpath;
  if (!(outpath = getenv("PROF_OUTPATH"))) {
    outpath = empty;
  }
#ifdef GZIP_PROFILE
#define EXT "out.gz"
#else
#define EXT "out"
#endif
  snprintf(buf, buflen, "%s/%s_%s.%s", outpath, basename, datebuf, EXT);
#undef EXT
}

// Will return correctly, or the whole program dies.
static FILE* out_file(const char *basename) {
  char outpath[OUTPATH_LEN];
  out_filename(outpath, OUTPATH_LEN, basename);
  FILE *out = fopen(outpath, "wb");
  DIE_NULL(out);
  return out;
}

// histogram format (binary)
// assume machine endianess
// Bytes | Data Type | Name
// -------------------------
// 0-3   | uint32    | nbbs
// 4-7   | uint32    | nthreads
// 8-11  | uint32    | type of profile (0 = active; 1 = working)
// 12-n  | uint32[]  | data, where n = 12 + 4*PER_THREAD_N
// n+1-  | char      | basic block names; each string is null-terminated

static void write_histogram() {
  fprintf(stderr, "[%lu] write_histogram\n", pthread_self());
  /*
  // Merge histograms before writing
  merge_hists();

#ifdef PROF_W
  char *name = "working";
#else
  char *name = "active";
#endif

  uint32_t nthreads = PROF_MAX_THREADS;
#ifdef PROF_W
  uint32_t type = 0;
#else
  uint32_t type = 1;
#endif

  FILE *f = out_file(name);
#ifdef GZIP_PROFILE
  gzFile *gf = gzdopen(fileno(f), "wb9");
#define FWRITE(buf, sz, ct, f) gzwrite((f), (buf), ((ct)*(sz)))
#define FH gf
#define PUTS(s, f) gzputs((f), (s))
#define PUTC(c, f) gzputc((f), (c))
#else
#define FWRITE fwrite
#define FH f
#define PUTS fputs
#define PUTC fputc
#endif

  FWRITE(&PROF_nbbs, 4, 1, FH);
  FWRITE(&nthreads,  4, 1, FH);
  FWRITE(&type,      4, 1, FH);
  FWRITE(PROF_hist, 4, PER_THREAD_N, FH);

  char **p = PROF_bbmap;
  while (*p) {
    PUTS(*p, f);
    PUTC(0, f); // reinsert a null termination
    ++p;
  }
#ifdef GZIP_PROFILE
  DIE_NZ(gzclose(FH));
#else
  DIE_NZ(fclose(FH));
#endif
*/
}

// ============================
// INSTRUMENTATION ENTRY POINTS
// ============================
void PROF_preblock() {
  fprintf(stderr, "[%lu] PROF_preblock\n", pthread_self());
#ifdef PROF_W
//   PROF_DEC_OFFSET();
#endif
}

void PROF_postblock() {
  fprintf(stderr, "[%lu] PROF_postblock\n", pthread_self());
#ifdef PROF_W
  // Dear Kui Tang:
  //
  // You are an idiot.
  //
  // FOR HOW LONG WAS THING DAMN THING PROF_DEC_OFFSET?!
//  PROF_INC_OFFSET();
#endif
}

void PROF_init_module() {
  fprintf(stderr, "[%lu] PROF_init_module\n", pthread_self());
  /*
  time_t start = time(NULL);
  DIE_NULL(localtime_r(&start, &start_time));

  // Remember to write stuff at program end
  DIE_NZ(atexit(write_histogram));

  // Set up the histograms
  PROF_hist = calloc(PROFILE_N, sizeof(uint32_t));
  DIE_NULL(PROF_hist);
  // Each thread initialization does add_and_fetch, so we start at index "-1".
  PROF_curr_thread_hist = PROF_hist;
  PROF_n_offset = 0;
  // We have one thread active, so we *must* have PROF_n_offset be 1, not 0.
  // This is because the main thread could be making blocking calls, which
  // would decrement PROF_n_offset. By starting at 1, we guarantee that
  // PROF_n_offset is always nonnegative.
  //
  // Umm, is this story even true?
  PROF_thread_hist = PROF_hist;

  // Handle thread-specific stuff
  DIE_NZ(pthread_key_create(&k_thread_exit, handle_thread_exit));

  fprintf(stderr, "[%lu] PROF_init_module completed\n", pthread_self());
  */
}

void PROF_init_thread() {
  fprintf(stderr, "[%lu] PROF_init_thread\n", pthread_self());
  /*
  PROF_INC_OFFSET();
  // Set the k_thread_exit to a dummy value (which we'll never use)
  DIE_NZ(pthread_setspecific(k_thread_exit, (void *) 1));
  PROF_thread_hist = PROF_INC_CURR();

  fprintf(stderr, "[%lu] PROF_init_thread completed; PROF_thread_hist = %x; PROF_curr_thread_hist = %x\n",
          pthread_self(), PROF_thread_hist, PROF_curr_thread_hist);
  */
}

void *PROF_shim_start_routine(void *wrapped_arg) {
  struct PROF_wrapper_t *wa = (struct PROF_wrapper_t*)(wrapped_arg);
  PROF_init_thread();
  void *ret = wa->start_routine(wa->arg);
  free(wrapped_arg);
  return ret;
}

void PROF_sample(int bb_id) {
  //PROF_thread_hist[PROF_n_offset + bb_id]++;
}

