#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include "rdtsc.h"

static uint64_t *vec;
static size_t vec_i;
static size_t vec_sz;

#define DIE_NULL(exp) do { if (!(exp)) { perror(#exp); exit(2); } } while (0)
#define DIE_NZ(exp) do { if ((exp)) { perror(#exp); exit(2); } } while (0)

void timer_tick(int signal, siginfo_t *info, void *p) {
  if (signal == SIGALRM) {
    if (vec_i >= vec_sz) {
      raise(SIGUSR1);
    } else {
      vec[vec_i] = rdtsc();
      vec_i++;
    }
  }
}

#define SQ(x) ((x)*(x))
void statistics(double *mean, double *std) {
  int i, diff_sz;
  diff_sz = vec_sz - 1;
  // compute adjacent differences
  for (i = 0; i < diff_sz; i++) { vec[i] = vec[i + 1] - vec[i]; }
  // compute mean and sd in batch
  *mean = 0;
  for (i = 0; i < diff_sz; i++) { *mean += vec[i]; }
  *mean /= diff_sz;

  // compute with Kahan compensated summation
  double s2 = 0, s3 = 0, resid;
  for (i = 0; i < diff_sz; i++) {
    resid = vec[i] - *mean;
    s2 += SQ(resid);
    s3 += resid;
  }
  double var = (s2 - SQ(s3) / diff_sz) / (diff_sz - 1);
  *std = sqrt(var);
}

void write_data(const char *path) {
  FILE *of = fopen(path, "w");
  DIE_NULL(of);

  for (int i = 0; i < vec_sz - 1; i++) {
    fprintf(of, "%lu\n", vec[i]);
  }

  fclose(of);
}

int main (int argc, char const* argv[]) {
  if (argc < 4) {
    fprintf(stderr, "Usage: %s NSAMPLES (suggested 100000) MICROSECS (sugg. 100) OUTFILE\n", argv[0]);
    return 1;
  }
  int us;
  vec_sz = atoi(argv[1]);
  us = atoi(argv[2]);
  if (vec_sz <= 0 || us <= 0) {
    fprintf(stderr, "Error: Invalid command line\n");
    return 1;
  }

  DIE_NULL(vec = malloc(vec_sz * sizeof(uint64_t)));

  sigset_t mask, old_mask;
  sigemptyset(&mask);
  sigaddset(&mask, SIGUSR1);

  DIE_NZ(sigprocmask(SIG_BLOCK, &mask, &old_mask));
  struct sigaction sa = { .sa_sigaction = timer_tick }, old_sa;
  DIE_NZ(sigaction(SIGALRM, &sa, &old_sa));

  struct timeval interval = { 0, us };
  struct itimerval itv = { interval, interval };
  DIE_NZ(setitimer(ITIMER_REAL, &itv, NULL));

  // Wait for SIGUSR1 to be raised
  siginfo_t info;
  while (sigwaitinfo(&mask, &info) < 0 && errno == EINTR)
    ;

  if (vec_i != vec_sz) {
    printf("Error: though vec_sz == %lu, vec_i == %lu\n", vec_sz, vec_i);
    return 2;
  }

  double mean, std;
  statistics(&mean, &std);
  int diff_sz = vec_sz - 1;
  printf("Of %d differences, mean = %g, sd = %g, sharpe = %g\n",
         diff_sz, mean, std, mean / std);

  // back of envelope normality test
  uint64_t min = UINT64_MAX, max = 0;
  int i;
  for (i = 0; i < diff_sz; i++) {
    if (vec[i] > max) { max = vec[i]; }
    else if (vec[i] < min) { min = vec[i]; }
  }
  printf("z_min, z_max = %g, %g\n", (min - mean) / std, (max - mean) / std);

  write_data(argv[3]);

  return 0;
}
